%% This is a sample code for dimensionality reduction using adaptive k-means
% supplementary code for RSS fingerprinting dataset size reduction using feature-wise adaptive k-means clustering
% by Lucie Klus, Darwin Quezada-Gaibor and Joaquin Torres-Sospedra

% The authors gratefully acknowledge funding from European Unions Horizon 2020 Research and Innovation programme 
% under the Marie Sklodowska Curie grant agreement No.813278
% (A-WEAR: A network for dynamicwearable applications with privacy constraints, http://www.a-wear.eu/).

%%
close all; clear all; clc;
% add paths to the folder with datasets, funtions etc.
addpath('databases');

%% load the desired dataset "database" in a structure format, with:
% trainingMacs - matrix of training fingerprints
% trainingLabels - matrix of training labels, including positioning
% information
% testMacs - matrix of testing fingerprints
% Testlabels - matrix of testing labels
datasetname = 'TUT2';
load(['databases/',datasetname,'.mat']);

%% initialize:choose data
X = database.trainingMacs; % set X as the training data

%% use k-means to reduce the dimensionality of data for 
% (case when all features are equivalent - all represent RSS values)
% calculate the number of unique values in the dataset
n_values_all = size(unique(X(:)),1); 
% calculate the number of clusters based on formula in the paper
K_all = ceil(n_values_all/sqrt(n_values_all)); 
% initiate clustering, X_cluster_all = sample cluster indeces,
% Centroids_all = centroid coordinates
[X_cluster_all, Centroids_all] = kmeans(X(:),K_all,'Distance','sqeuclidean',...
   'Replicates',50); 
% calculate the number of samples in each cluster
for j = 1:K_all
    samplecount_all(j) = sum(X_cluster_all==j); 
end
% create the transformed dataset for validation, in which each entry is represented by the
% cluster coordinate
X_cluster_all_coord = Centroids_all(X_cluster_all);
X_cluster_all_coord = reshape(X_cluster_all_coord,size(X));

%% calculate and print the reconstruction mean squared error
MSE_all = immse(X,X_cluster_all_coord);
fprintf(['Current dataset: ',datasetname,'\n'])
fprintf('Results for global feature clustering: \n    Number of clusters: %1.0f \n    Training dataset error: %1.5f MSE \n', K_all, MSE_all);

%% cluster the training data (without per-sample centroid updates)
X_test = database.testMacs;
X_test_exp = kron(X_test(:),ones(1,K_all));         % expand matrix dimensions
distances = sqrt((X_test_exp-Centroids_all').^2);   % calculate euclidean distances
[mindist, minind] = min(distances,[],2);            % find closest centroids
X_test_rec = Centroids_all(minind);                 % substitute values for centroid coordinates
X_test_rec = reshape(X_test_rec,size(X_test));      % reshape back to original matrix dimensions
X_test_ind = reshape(minind,size(X_test));          % create matrix of cluster indexes
% calculate and print the reconstruction mean squared error for the testset
MSE_test = immse(X_test,X_test_rec);
fprintf('    Test dataset error: %1.5f MSE (without centroid updates)\n', MSE_test);

%% cluster the training data (with centroids shifts in per-sample loop)
Centroids_all_shifted = Centroids_all;                  % create the array for recording shifted coordinates
samplecount_all_shifted = samplecount_all;              % create the array for recording shifted coordinate counts
X_test_rec_update = zeros(size(X_test));            
X_test_ind_update = zeros(size(X_test));
% initiate the per-sample loop
for sample = 1:size(X_test,1)
newsample = X_test(sample,:)';                           % currently considered sample
newsample_exp = kron(newsample,ones(1,K_all));           % expand matrix dimensions
distances2 = sqrt((newsample_exp-Centroids_all').^2);    % calculate euclidean distances
[mindist2, minind2] = min(distances2,[],2);              % find closest centroid for each feature in the sample
% shift centroid coordinates according to the sample based on the paper
for feature = 1:size(newsample,1)                        
Centroids_all_shifted(minind2(feature)) = ...
    (Centroids_all_shifted(minind2(feature))*samplecount_all_shifted(minind2(feature))+newsample(feature))/...
    (samplecount_all_shifted(minind2(feature))+1);
samplecount_all_shifted(minind2(feature)) = ...
    samplecount_all_shifted(minind2(feature))+1;
end
X_test_ind_update(sample,:) = minind2';
end
X_test_rec_update = Centroids_all_shifted(X_test_ind_update);       % create matrix of cluster coordinates
% calculate and print the reconstruction mean squared error for the testset
% with centroid updates
MSE_test2 = immse(X_test,X_test_rec_update);
fprintf('    Test dataset error: %1.5f MSE (with centroid updates)\n', MSE_test2);

%% validate the method for fingerprinting localization
conf.k = 10; % number of closest considered samples for K-NN

%% evaluate fingerprinting accuracy of the original dataset
KNNmodel_rawX = fitcknn(database.trainingMacs,database.trainingLabels(:,1),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
KNNmodel_rawY = fitcknn(database.trainingMacs,database.trainingLabels(:,2),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
KNNmodel_rawZ = fitcknn(database.trainingMacs,database.trainingLabels(:,3),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
estimatedlabelsX = predict(KNNmodel_rawX,database.testMacs);
estimatedlabelsY = predict(KNNmodel_rawY,database.testMacs);
estimatedlabelsZ = predict(KNNmodel_rawZ,database.testMacs);
estimatedlabels = [estimatedlabelsX,estimatedlabelsY,estimatedlabelsZ];
[error2D,error3D] = error_estimation(estimatedlabels,database.testLabels);
floor_hit = sum(estimatedlabelsZ == database.testLabels(:,3));

%% evaluate fingerprinting accuracy of the dataset after clustering
KNNmodel_allX = fitcknn(X_cluster_all_coord,database.trainingLabels(:,1),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
KNNmodel_allY = fitcknn(X_cluster_all_coord,database.trainingLabels(:,2),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
KNNmodel_allZ = fitcknn(X_cluster_all_coord,database.trainingLabels(:,3),'NumNeighbors',conf.k,...
    'NSMethod','exhaustive','Distance','euclidean','Standardize',0);
estimatedlabelsXall = predict(KNNmodel_allX,X_test_rec_update);
estimatedlabelsYall = predict(KNNmodel_allY,X_test_rec_update);
estimatedlabelsZall = predict(KNNmodel_allZ,X_test_rec_update);
estimatedlabelsall = [estimatedlabelsXall,estimatedlabelsYall,estimatedlabelsZall];
[error2D2,error3D2] = error_estimation(estimatedlabelsall,database.testLabels);
floor_hit2 = sum(estimatedlabelsZall == database.testLabels(:,3));

fprintf('Results for %2d-NN classifier\n', conf.k)
fprintf('Before the data compression\n')
fprintf('    Mean Positioning Error 2D   : %3.7f\n',error2D); % In meters X,Y
fprintf('    Mean Positioning Error 3D   : %3.7f\n',error3D); % In meters X,Y
fprintf('    Floor hit                   : %3.7f\n',floor_hit);
fprintf('After the data compression\n')
fprintf('    Mean Positioning Error 2D   : %3.7f\n',error2D2); % In meters X,Y
fprintf('    Mean Positioning Error 3D   : %3.7f\n',error3D2); % In meters X,Y
fprintf('    Floor hit                   : %3.7f\n',floor_hit2);
fprintf('    Normalized 3D error:        : %3.7f\n',error3D2/error3D);

%% dataset size reduction:
%% 64 bit representation
% % original measurement dataset size: (double = 8 bytes = 64 bits, real-valued measurements)
% size_Orig = 64*(size(X(:),1)+size(X_test(:),1));
% % compressed measurement dataset size:
% size_Comp = ceil(log2(K_all))*(size(X(:),1)+size(X_test(:),1));
% % additional overhead:
% size_Overhead = 64*(size(Centroids_all_shifted(:),1)+size(samplecount_all_shifted(:),1));

%% 7 bit representation
% original measurement dataset size: (integer values of RSS data (up to 128 unique values))
size_Orig = 7*(size(X(:),1)+size(X_test(:),1));
% compressed measurement dataset size:
size_Comp = ceil(log2(K_all))*(size(X(:),1)+size(X_test(:),1));
% additional overhead:
size_Overhead = 64*(size(Centroids_all_shifted(:),1)+size(samplecount_all_shifted(:),1));

%%

fprintf('Compression efficiency:\n')
fprintf('    Original dataset size   : %3.0f bits\n',size_Orig);
fprintf('    Compressed dataset size   : %3.0f bits\n',size_Comp);
fprintf('    Additional required overhead   : %3.0f bits\n',size_Overhead);
fprintf('    Compressed dataset size   : %3.5f%% of the original data\n',(size_Overhead+size_Comp)/size_Orig*100);
fprintf('    CR                        : %3.5f \n', size_Orig/(size_Overhead+size_Comp));
fprintf('    Size of original          : %3.0f \n', size_Orig);
fprintf('    Size of compressed        : %3.0f + %3.0f = %3.0f \n',size_Comp, size_Overhead,(size_Overhead+size_Comp));
fprintf('    Number of samples         : %3.0f \n',size(X,1)+size(X_test,1));

% function for calculating 2D and 3D positioning error
function [error2D,error3D] = error_estimation(estimatedlabels, testLabels)
    rowEstimatedLabels = size(estimatedlabels,1);
    dist_3D = 0;
    dist_2D = 0;
    for i = 1:rowEstimatedLabels
        dist_3D = dist_3D + sqrt(sum((estimatedlabels(i,:)-testLabels(i,1:3)).^2));
        dist_2D = dist_2D + sqrt(sum((estimatedlabels(i,1:2)-testLabels(i,1:2)).^2));
    end
    error2D = dist_2D/rowEstimatedLabels;
    error3D = dist_3D/rowEstimatedLabels;
end